@@ -4,14 +4,14 @@ import os
 
 import tensorflow as tf
 from tensorflow.keras.models import load_model
-from tensorflow.keras.layers import Dense, Input
-from tensorflow.keras.models import Model
+
 from keras import backend
 
 import pandas as pd
 import numpy as np
 
 from cfbench.cfg import OHConverter
+from cfbench.cfg.common import nn_ohe
 
 from cfbench.dataset_data.constants.var_types import VAR_TYPES
 
@@ -29,38 +29,25 @@ except AttributeError:
 tf.compat.v1.disable_v2_behavior()  # disable TF2 behaviour as alibi code still relies on TF1 constructs
 
 
-def _nn_ohe(input_shape, hidden_layers_ws, output_number):
-    x_in = Input(shape=(input_shape,))
-    x = Dense(hidden_layers_ws, activation='relu')(x_in)
-    x_out = Dense(2, activation='softmax')(x)
-    if output_number == 1:
-        x_bin = Dense(1, activation='linear')(x_out)
-        nn = Model(inputs=x_in, outputs=x_bin)
-    if output_number == 2:
-        nn = Model(inputs=x_in, outputs=x_out)
-
-    nn.compile(loss='categorical_crossentropy', optimizer='adam', metrics=['accuracy'])
-
-    return nn
 
 
 class BenchmarkGenerator:
 
     def __init__(
             self,
-            framework_name,
             output_number,
             ds_id_test,
             disable_gpu):
-        self.framework_name = framework_name
         self.output_number = output_number
         self.ds_id_tes = ds_id_test
 
         self.ds_idx = 0
         self.current_dsName = None
         self.factual_class = 0
+        self.current_factual_class = 0
         self.factual_idx = 0
         self.total_factuals = None
+        self.save_results = False
 
         if disable_gpu:
             # Disable GPU
@@ -96,12 +83,13 @@ class BenchmarkGenerator:
 
         dsName = list(VAR_TYPES.keys())[self.ds_id_tes[self.ds_idx]]
 
-        if dsName == self.current_dsName:
+        if dsName == self.current_dsName and self.current_factual_class == self.factual_class:
             # Go to next row
             self.factual_idx += 1
         else:
-            # Assign new dataset name
+            # Assign new dataset name and factual class
             self.current_dsName = dsName
+            self.current_factual_class = self.factual_class
 
             # Load feature type specifications
             self.cat_feats = VAR_TYPES[dsName]['categorical']
@@ -169,7 +157,7 @@ class BenchmarkGenerator:
             input_shape = model_keras.get_weights()[0].shape[0]
             hidden_layers_ws = model_keras.get_weights()[0].shape[1]
 
-            self.adapted_nn = _nn_ohe(input_shape, hidden_layers_ws, self.output_number)
+            self.adapted_nn = nn_ohe(input_shape, hidden_layers_ws, self.output_number)
             self.adapted_nn.build(input_shape=input_shape)
 
             self.adapted_nn.layers[1].set_weights(model_keras.layers[0].get_weights())
@@ -224,7 +212,10 @@ class BenchmarkGenerator:
             'dsname': self.current_dsName
         }
 
-    def cf_evaluator(self, cf_out: list, verbose: bool = False):
+    def cf_evaluator(
+            self, cf_out: list, algorithm_name: str, cf_generation_time: float = None,
+            verbose: bool = False, save_results: bool = False):
+        self.save_results = save_results
         # Verify if counterfactual (cf) is a list, the output MUST a list
         try:
             assert type(cf_out) == list
@@ -270,8 +261,31 @@ class BenchmarkGenerator:
                     logging.info(f'Counterfactual found!\nFactual class:{factual_class}\nCF class:{cf_class}\n')
                 else:
                     logging.info(f'Failed counterfactual!\nFactual class:{factual_class}\nCF class:{cf_class}\n')
-            return True, cf_out_oh
+            cf_found = True
+            processed_cf = cf_out_oh
         else:
             if verbose:
                 logging.info(f'No returned counterfactual candidate!\n')
-            return False, [np.NaN] * (len(self.factual_oh))
+            cf_found = False
+            processed_cf = [np.NaN] * (len(self.factual_oh))
+
+        # Save result in local folder
+        if save_results:
+            if not os.path.exists('./cfbench_results/'):
+                os.mkdir('./cfbench_results/')
+            save_df = [{
+                'algorithm_name': algorithm_name,
+                'dataset_name': self.current_dsName,
+                'factual_idx': self.factual_idx,
+                'factual_class': self.factual_class,
+                'cf_found': cf_found,
+                'cf': processed_cf,
+                'factual': self.factual_oh,
+                'cf_generation_time': cf_generation_time
+            }]
+            pd.to_pickle(
+                save_df,
+                f'./cfbench_results/'
+                f'{algorithm_name}_{self.current_dsName}_{self.factual_class}_{self.factual_idx}.pkl')
+
+        return cf_found, processed_cf
